;       AX.25 rx/tx routine                        N0QBH   2000
;       10 Mhz XTAL
;       Relays valid compressed MIC-E/PIC-E data
;       VISUAL 1 on after rx two flags
;       SW 1 toggles valid MIC-E filter
;       My serial routine = data out A:1
;       4/01/00
;____________________________________________________________
	LIST	P=PIC16F84
	include	"P16F84.INC"
	__CONFIG _HS_OSC & _CP_OFF & _PWRTE_ON & _WDT_OFF
;------------------------------------------------------------
;               Port Set-up
;           0   1   2   3   4   5   6   7       bit
; port  A   Out Out Out In  In  X   X   X
;       B   Out Out Out In  In  Out Out Out
;____________________________________________________________
	ORG	0			; start code
	goto	main
;------------------------------------------------------------
my_call_sign
        addwf   PCL,F
        dt      "N0QBH ",0              ;6 chars only
;------------------------------------------------------------
        cblock  0x0c
                flag_count              ;flag counter  (0c)
                byte_count              ;delay counter
                store_count             ;bytes stored cnt
                serial_count            ;serial out counter
                delay_count             ;delay counter
                bit_stuff_count         ;B_S counter
                bit_count               ;bit counter
                send_reg                ;serial data out
                FLAGS                   ;various flags
                snap_shot               ;comparisons
                byte_in                 ;data byte rx'd
                AX25_crc_hi             ;upper byte CRC word
                AX25_crc_lo             ;lower byte CRC word
                next_to_last_byte       ;2nd from last byte
                last_byte               ;last byte rxed  (1a)
        endc
;------------------------------------------------------------
#define SERIAL_DATA_OUT         PORTA,1
#define PTT_OUT                 PORTA,2
#define MODE_SWITCH             PORTA,4
#define VISUAL_ONE              PORTB,0
#define MODE1_OUT               PORTB,1
#define MODE0_OUT               PORTB,2
#define PKT_DATA_OUT            PORTB,5
#define DATA_FLAG               FLAGS,1
#define SHIFTED_DATA            FLAGS,2
#define SELECT_MODE             FLAGS,3
;------------------------------------------------------------
;
main
        clrwdt
        clrf    TMR0
        bsf     STATUS,RP0              ;bank 1
        movlw   b'11111000'             ;A:3,:4 inputs
        movwf   TRISA                   ;set up port A I/O
        movlw   b'00011000'             ;B:3,4 input
        movwf   TRISB                   ;set up port B I/O
        movlw   b'11010011'
        movwf   OPTION_REG              ;TMR0 /16
        bcf     STATUS,RP0              ;bank 0
        clrf    INTCON                  ;clear INT reg
        bcf     PTT_OUT                 ;unkey
        bcf     SERIAL_DATA_OUT         ;stop set
        bsf     MODE0_OUT
        bsf     MODE1_OUT               ;MX 614 init
        movlw   0x0a                    ;10 loops
        call    adjustable_delay        ;wait 2 mS
        bcf     MODE0_OUT               ;RX mode
        movlw   0x64                    ;100 loops
        call    adjustable_delay        ;20 mS delay
        call    wake_up_wink            ;flash LED
rx_set_up
        bcf     MODE0_OUT               ;MX614 setup
        bsf     MODE1_OUT               ;1200 rx
        bcf     PKT_DATA_OUT            ;data out clear
        movlw   0x14                    ;20 loops
        call    adjustable_delay        ;wait 4 mS
;------------------------------------------------------------
house_clean                             ;new packet
        clrf    snap_shot               ;reset comparisons
        clrf    byte_in                 ;reset bytes rx count
        clrf    bit_stuff_count         ;reset b.s. count
        clrf    store_count             ;reset count out
        clrf    byte_count              ;reset byte count
        clrf    flag_count              ;reset flag count
        clrf    bit_count               ;reset bit count
        clrf    send_reg                ;reset serial out
        clrf    FLAGS                   ;reset flags
        btfsc   MODE_SWITCH             ;which mode ?
        bsf     SELECT_MODE             ;rx only mode
        movlw   0xff                    ;relay mode
        movwf   AX25_crc_hi             ;reset CRC calc
        movwf   AX25_crc_lo
        movlw   0x20
        movwf   FSR                     ;reset FSR
        bcf     VISUAL_ONE              ;#1 LED off
        movlw   0x7e                    ;130 counts (832 uS)
        movwf   TMR0                    ;til T0IF set
        movfw   PORTB                   ;reset RBI
        clrf    INTCON                  ;set up TMR0, RBI
find_start_flag
        clrc
        call    recv_one_bit            ;check for toggle
        skpnz
        setc                            ;no toggle (1)
        rrf     byte_in,F               ;toggle  (0)
        movfw   byte_in                 ;byte_in to W
        xorlw   0x7e                    ;is it a flag ?
        skpz
        goto    find_start_flag         ;no
start_flag_loop                         ;yes, look again
        call    recv_one_bit
        skpnz                           ;see a toggle ?
        goto    find_start_flag         ;no, go back
        movlw   0x06                    ;yes, must see 6 1's
        movwf   store_count
long_ones
        call    recv_one_bit            ;look for 6
        skpz                            ;1's in arow
        goto    find_start_flag         ;saw a 0
        decfsz  store_count,F           ;saw a 1
        goto    long_ones
        call    recv_one_bit            ;look for end 0
        skpnz
        goto    find_start_flag         ;saw a 1, restart
        bsf     VISUAL_ONE              ;flag, LED on
;------------------------------------------------------------
;
get_next_byte
        clrf    byte_in                 ;clear input byte
        movlw   0x08
        movwf   bit_count               ;get 8 bits
get_next_bit
        call    recv_one_bit            ;get 1 bit
        skpz                            ;skip if W is 0
        goto    zero_in                 ;0 rx'd
        incf    bit_stuff_count,F       ;1, inc counter
        setc                            ;set carry
roll_em_in
        rrf     byte_in,F               ;bit stored
        movfw   bit_stuff_count         ;B_S_F to W
        xorlw   0x05                    ;W = 5 ?
        skpnz
        goto    check_stuff             ;yes
count_em_out                            ;no
        decfsz  bit_count,F             ;8 bits ?
        goto    get_next_bit            ;no
        goto    CRC_save_store          ;yes
zero_in
        clrf    bit_stuff_count         ;0 seen, clear bsc
        clrc                            ;clear carry
        goto    roll_em_in              ;get another bit
;------------------------------------------------------------
; Bit stuff decode routine
;
check_stuff                             ;saw 5 1's in a row
        call    recv_one_bit            ;wait 1 bit
        skpnz                           ;if data=0, skip next
        goto    sixth_one               ;unchanged, flag ?
        clrf    bit_stuff_count         ;tog seen, clear count
        goto    count_em_out            ;don't store stuffed bit
sixth_one
        call    recv_one_bit            ;get next bit
        skpnz
        goto    house_clean             ;7th 1, reset
        movfw   bit_count               ;is flag aligned
        xorlw   0x03                    ;in the byte ?
        skpz
        goto    house_clean             ;no, reset
        clrf    bit_stuff_count         ;yes, clear B_S_C
        btfss   DATA_FLAG               ;header/tail flag ?
        goto    get_next_byte           ;header flag, get next
        movfw   byte_count              ;tail flag, end of packet?
        sublw   0x0f                    ;how many bytes rx'd ?
        skpnc
        goto    house_clean             ;< 17 bytes rx'd
        goto    packet_end              ;it's long enough
;
CRC_save_store
        btfss   DATA_FLAG               ;test data flag
        bsf     DATA_FLAG               ;set data flag
        incf    byte_count,F            ;count bytes
        movfw   byte_count
        sublw   0xa0                    ;how many bytes ?
        skpc
        goto    house_clean             ; 160 bytes, reset
        movfw   byte_count              ;<160 bytes, OK
        sublw   0x02                    ;CRC calc is done on
        skpc                            ;3rd from last byte
        call    AX25_CRC_calc           ;rx'd
        movfw   last_byte               ;last byte to
        movwf   next_to_last_byte       ;2nd to last byte
        movfw   byte_in                 ;get new data byte
        movwf   last_byte               ;store in last byte
        goto    get_next_byte           ;get next
;
;------------------------------------------------------------
; Invert CRC data and compare to value sent
packet_end
        movlw   0xff
        xorwf   AX25_crc_lo,F           ;invert CRC lo
        xorwf   AX25_crc_hi,F           ;invert CRC hi
        movfw   next_to_last_byte
        xorwf   AX25_crc_lo,W           ;CRC lo bits match ?
        skpz                            ;yes, skip next
        goto    house_clean             ;mismatch, reset
        movfw   last_byte
        xorwf   AX25_crc_hi,W           ;CRC hi bytes match ?
        skpz                            ;yes, skip next
        goto    house_clean             ;mismatch, reset
        movfw   FSR                     ;
        movwf   store_count             ;end of data loc
        decf    store_count,F           ;store count -1
;------------------------------------------------------------
        call    find_E_O_A
        btfss   SHIFTED_DATA
        goto    house_clean             ;no find, reset
        btfss   SELECT_MODE             ;set = filtered
        goto    transmit_set_up         ;clear = all
;------------------------------------------------------------
        movfw   byte_in                 ;get E_O_A loc
        movwf   FSR                     ;start here
;------------------------------------------------------------
        incf    FSR,F                   ;start of PID
        movfw   INDF
        xorlw   0x03                    ; = 03h ?
        skpz
        goto    house_clean             ;no
        incf    FSR,F                   ;yes, look at next
        movfw   INDF
        xorlw   0xf0                    ;= f0h ?
        skpz
        goto    house_clean             ;no
        incf    FSR,F                   ;yes, look at next
        movfw   INDF
        xorlw   0x60                    ;= 60h ?
        skpz
        goto    house_clean             ;no
;------------------------------------------------------------
transmit_set_up
        movlw   0xff
        movwf   AX25_crc_hi             ;reset CRC calc
        movwf   AX25_crc_lo             ;bytes to FF
        bcf     MODE1_OUT
        bsf     MODE0_OUT               ;MX614 tx  mode
        call    onebit_delay            ;wait 208 uS
        clrf    snap_shot               ;now used for toggle
        clrf    bit_stuff_count         ;B_S_C for tx
        bcf     PKT_DATA_OUT            ;start tone
        bsf     PTT_OUT                 ;key radio
        movlw   0x7e                    ;832 uS til T0IF set
        movwf   TMR0
        bcf     INTCON,T0IF             ;set up timer
        movlw   0x28
        movwf   flag_count              ;40 flag delay
        call    tx_flags                ;send flags
        movlw   0x20
        movwf   FSR                     ;reset FSR pointer
;
send_pkt_byte
        movfw   INDF
        movwf   send_reg                ;INDF to send_reg
        call    reg_packet_out          ;send byte
        incf    FSR,W                   ;inc FSR to W
        xorlw   0x2e                    ;15th byte ?
        skpnz
        call    insert_callsign         ;yes, insert call
        incf    FSR,W                   ;FSR to W
        subwf   store_count,W           ;same ?
        skpc
        goto    send_crc_word           ;yes, done
        incf    FSR,F                   ;no, inc FSR
        goto    send_pkt_byte           ;loop
send_crc_word
        movlw   0xff
        xorwf   AX25_crc_hi,W           ;invert both
        movwf   next_to_last_byte       ;CRC bytes
        movlw   0xff
        xorwf   AX25_crc_lo,W
        movwf   send_reg
        call    reg_packet_out          ;send CRC lo byte
        movfw   next_to_last_byte
        movwf   send_reg
        call    reg_packet_out          ;send CRC hi byte
        movlw   0x02
        movwf   flag_count
        call    tx_flags                ;send 2 end flags
        bcf     PTT_OUT                 ;unkey radio
;
;-----------------------------------------------------------
;       Serial out routine
out_at_4k8
;
        movlw   0x20
        movwf   FSR                     ;reset FSR
read_buffer
        movfw   FSR                     ;FSR to W
        subwf   byte_in,W               ;
        skpc
        goto    no_shift_byte           ;no more shift
        movfw   INDF                    ;shift this byte
        movwf   send_reg                ;byte to send_reg
        clrc
        rrf     send_reg,F              ;right shift byte
        goto    bypass_load             ;send it
no_shift_byte
        movfw   INDF
        movwf   send_reg
bypass_load
        call    serial_out              ;send the byte
        incf    FSR,W
        subwf   store_count,W           ;all bytes sent ?
        skpc
        goto    carrage_return          ;yes
        incf    FSR,F
        goto    read_buffer             ;no
carrage_return
        movlw   0x0d                    ;yes, send ending
        movwf   send_reg
        call    serial_out              ;send <CR>
        movlw   0x0a
        movwf   send_reg
        call    serial_out              ;send <LF>
        goto    rx_set_up               ;done, start over
;------------------------------------------------------------
;       Subroutines
;------------------------------------------------------------
wake_up_wink
        movlw   0x10
        movwf   byte_count
        bsf     VISUAL_ONE              ;LED on
wake_up_wink_loop
        movlw   0xff
        call    adjustable_delay        ;wait
        call    adjustable_delay
        bcf     VISUAL_ONE
        movlw   0xff
        call    adjustable_delay
        call    adjustable_delay
        bsf     VISUAL_ONE
        decfsz  byte_count,F
        goto    wake_up_wink_loop
        bcf     VISUAL_ONE
        return
;------------------------------------------------------------
insert_callsign
        bcf     INDF,0                  ;clear any E_O_A bit
        clrf    byte_count
insert_call_loop
        movfw   byte_count
        call    my_call_sign            ;look up table
        movwf   send_reg
        xorlw   0x00
        skpnz
        goto    set_f_flag              ;last byte
        clrc
        rlf     send_reg,F              ;left shift data
        call    reg_packet_out          ;send byte
        incf    byte_count,F            ;inc lookup offset
        goto    insert_call_loop
set_f_flag
        movlw   0x26
        xorwf   byte_in,W               ;test for short addr
        skpnz
        bsf     send_reg,0              ;sets E_O_D flag
        bsf     send_reg,7              ;sets F bit
        call    reg_packet_out
        return
;------------------------------------------------------------
;               End of address finder
find_E_O_A
        bcf     SHIFTED_DATA            ;clear flag
        movlw   0x2d                    ;1st E_O_A loc
        movwf   FSR                     ;look for E_O_A flag
address_list_end                        ;at each SSID loc
        btfsc   INDF,0                  ;test byte
        goto    set_flag_store_loc      ;E_O_A found
        movlw   0x07                    ;look at next SSID
        addwf   FSR,W                   ;add 7 bytes, test
        sublw   0x42
        skpc
        return                          ;> 3 digis, no flag
        movlw   0x07                    ;inc by 7
        addwf   FSR,F
        goto    address_list_end
set_flag_store_loc
        bsf     SHIFTED_DATA            ;set flag
        movfw   FSR
        movwf   byte_in                 ;store E_O_A loc
        return
;------------------------------------------------------------
;               Assumes byte is in send_reg
reg_packet_out
        movlw   0x08
        movwf   bit_count               ;8 bits
rotor_out
        rrf     send_reg,W              ;calc CRC
        call    CRC_core
        rrf     send_reg,F              ;get bit to send
        skpc
        goto    zero_out                ;it's a zero
        incf    bit_stuff_count,F       ; B_S_C +1
        call    tx_one_bit              ;send a 1
        movfw   bit_stuff_count         ;check B_S_C
        xorlw   0x05
        skpnz                           ;B_S_C = 5 ?
        goto    zero_out                ;yes
count_the_bits
        decfsz  bit_count,F             ;8 bits sent ?
        goto    rotor_out               ;no
        return                          ;yes
zero_out
        clrf    bit_stuff_count         ;reset B_S_C
        incf    snap_shot,F             ;toggle tone
        call    tx_one_bit              ;sends a 0
        goto    count_the_bits          ;loop back
;------------------------------------------------------------
;               Sends 01111110 flag_count times
tx_flags
        movlw   0x06
        movwf   bit_count               ;6 ones
        incf    snap_shot,F
        call    tx_one_bit              ;send a 0
flag_twist
        call    tx_one_bit              ;send 6 1's
        decfsz  bit_count,F             ;6 sent ?
        goto    flag_twist              ;no
        incf    snap_shot,F             ;yes
        call    tx_one_bit              ;send a 0
        decfsz  flag_count,F            ;all flags sent ?
        goto    tx_flags                ;no
        return                          ;yes, end of flags
;------------------------------------------------------------
AX25_CRC_calc
        incf    FSR,W                   ;inc FSR
        sublw   0x4f                    ;rx data buffer full?
        skpc
        goto    load_CRC_bits           ;yes, skip store
        movfw   next_to_last_byte       ;
        movwf   INDF                    ;store at offset
        incf    FSR,F
load_CRC_bits
        movlw   0x08                    ;calc byte's CRC
        movwf   bit_count               ;8 loops
CRC_head
        rrf     next_to_last_byte,F     ;get bit for CRC
        call    CRC_core                ;do the calc
        decfsz  bit_count,F             ;8 bits done ?
        goto    CRC_head                ;no
        return                          ;yes
;------------------------------------------------------------
; Assumes carry contains data bit
CRC_core
        movlw   1
        skpnc
        xorwf   AX25_crc_lo, F          ;calculates CRC
        clrc                            ;as per AX25 spec
        rrf     AX25_crc_hi, F
        rrf     AX25_crc_lo, F
        skpc
        return
        movlw   H'08'
        xorwf   AX25_crc_lo, F
        movlw   H'84'
        xorwf   AX25_crc_hi, F
        return
;------------------------------------------------------------
;       Packet bit receive routine
;       Returns with 0 or 1 in Zero Flag
recv_one_bit                            ;delay 833 uS
        btfsc   INTCON,RBIF
        goto    bit_edge                ;saw tone toggle
        btfss   INTCON,T0IF
        goto    recv_one_bit            ;not timed out
        movlw   0x7e                    ;timed out
        movwf   TMR0                    ;reset timer
        bcf     INTCON,T0IF             ;1 byte long
        clrw                            ;Z flag set
        return
bit_edge
        movlw   0x3d                    ;load 1 1/2 bits
        movwf   TMR0
        movfw   PORTB                   ;reset port change
        bcf     INTCON,RBIF             ;clear RBIF
        bcf     INTCON,T0IF
        movfw   TMR0                    ;Z flag clear
        return
;------------------------------------------------------------
;               uses snapshot reg to toggle
;
tx_one_bit                              ;832 uS delay
        btfss   INTCON,T0IF
        goto    tx_one_bit
        movfw   snap_shot               ;snap_shot to W
        andlw   0x01                    ;mask upper 7 bits
        skpz                            ;0 or 1 ?
        goto    lo_tone                 ;1, goto lo_tone
        bcf     PKT_DATA_OUT            ;0, send 2200 Hz
        goto    reset_the_timer
lo_tone
        bsf     PKT_DATA_OUT            ;send 1200 Hz
reset_the_timer                         ;one, no change
        movlw   0x7e                    ;824uS
        movwf   TMR0
        bcf     INTCON,T0IF             ;clr timeout flag
        return
;------------------------------------------------------------
;------------------------------------------------------------
;               4800,8,N,1      serial out routine
;
serial_out
        movlw   0x08                    ;init shift counter
        movwf   serial_count
        bsf     SERIAL_DATA_OUT         ;start bit set
        call    onebit_delay
        comf    send_reg,F              ;invert data
        clrc
serial_out_loop
        rrf     send_reg,F              ;rotate bit 0 into carry
        skpnc                           ;test carry flag
        goto    hbit                    ;carry = 1
        bcf     SERIAL_DATA_OUT         ;bit is clear, output 0
        goto    time2_roll
hbit    bsf     SERIAL_DATA_OUT         ;bit is set, output 1
time2_roll
        call    onebit_delay
        decfsz  serial_count,F          ;shifted 8?
        goto    serial_out_loop         ;no
        bcf     SERIAL_DATA_OUT         ;yes, output stopbit
        call    onebit_delay
        return                          ;done
;------------------------------------------------------------
;               10 Mhz version
onebit_delay                            ;208 uS delay (4k8)
        movlw   0xad                    ;173 loops @ 1.2 uS
        movwf   flag_count
onebit_loop
        decfsz  flag_count,F
        goto    onebit_loop
        return
;------------------------------------------------------------
adjustable_delay                           ;W contains length
        movwf   delay_count
adju_mS_loop
        call    onebit_delay
        decfsz  delay_count,F
        goto    adju_mS_loop
        return
;------------------------------------------------------------
        END
